"""
dnd-character is a Python package for integrating DnD characters
into external applications.
"""


__author__ = "Brianna Rainey"
__copyright__ = "Copyright 2019-2020 Brianna Rainey & Markis Cook"
__credits__ = ["Brianna Rainey (Current Programmer)", "Markis Cook (Original Creator)"]
__license__ = "EPL-2.0"
__version__ = "20.11.01"
__maintainer__ = "Brianna Rainey"


from .character import Character



from random import SystemRandom


class Roll(object):
    """Creates roll values

    The Roll object creates roll values for dice rolls.  Random values are created from
    SystemRandom (using system entropy) to create a better pseudorandom number.

    Args:
            min (int): the minimum number to roll
            max (int): the maximum number to roll

    Returns:
            This object returns nothing and instead sets the value of self.value
            to be called at a later time.
    """

    def __init__(self, min: int, max: int, modifier: int = 0):
        self.min = min
        self.max = max + 1
        self.modifier = modifier
        self.dice()

    def dice(self):
        c = SystemRandom()
        self.value = c.randrange(self.min, self.max) + self.modifier


class RollStats(object):
    def __init__(self, method: str = None):
        self.method = method if method is None else method.lower()

    def stat_roller(self):
        if self.method in [None, "standard", "4d6d1", "4d6dl"]:
            # default method
            statNumber = 0
            self.rollList = []
            for i in range(4):
                dice = Roll(1, 6)
                statNumber += dice.value
                self.rollList.append(dice.value)
            statNumber -= min(self.rollList)
            self.rollList.remove(min(self.rollList))
            self.statroll = int(statNumber)

        elif self.method == "3d6" or "classic":
            # Roll 4D6 drop lowerest method
            statNumber = 0
            self.rollList = []
            for i in range(3):
                dice = Roll(1, 6)
                statNumber += dice.value
                self.rollList.append(dice.value)
            self.statroll = int(statNumber)

        elif self.method == "heroic" or "2d6+6":
            # Roll 2d6 and add 6 to that number
            statNumber = 0
            self.rollList = []
            for i in range(2):
                dice = Roll(1, 6)
                statNumber += dice.value
                self.rollList.append(dice.value)
            self.statroll = int(statNumber + 6)

        else:
            raise "Accepted values are, 'standard','classic', and 'heroic' roll methods."

        return self.statroll



"""
A cached function that gets SRD data from a DND 5e REST API
"""
import requests
import json
from os import environ, walk, path, remove
from functools import wraps
import logging


LOG = logging.getLogger(__package__)
LOG.setLevel(logging.DEBUG)


try:
    JSON_CACHE = f"{path.dirname(path.abspath(__file__))}/json_cache"
except Exception as e:
    LOG.error(f"Entire JSON cache failed to load: {str(e)}")


def cached_json(func):
    """
    Decorator which caches REST get requests into a dict
    and saves them to JSON files so the cache can be reloaded
    """
    func.cache = {}
    for dirname, __, files in walk(JSON_CACHE):
        for fp in files:
            if path.splitext(fp)[1] != ".json":
                continue
            try:
                with open(f"{dirname}/{fp}", "r") as f:
                    data = json.load(f)
                func.cache[f"/{fp.replace('_', '/')[:-5]}"] = data
            except json.decoder.JSONDecodeError as e:
                LOG.error(f"{fp} failed to load: {str(e)}")
                remove(f"{dirname}/{fp}")

    def outer_wrapper(func):
        def inner_wrapper(uri):
            try:
                return func.cache[uri]
            except KeyError:
                LOG.debug(f"Uncached URI: {uri}")
                func.cache[uri] = result = func(uri)
                fp = f"{JSON_CACHE}/{uri[1:].replace('/', '_')}.json"
                with open(fp, "w") as f:
                    f.write(json.dumps(result))
                return result

        return inner_wrapper

    return outer_wrapper(func)


def __SRD_API_CALL():
    """
    Closure for API calls
    """
    SRD_API = environ.get("SRD_API", "http://dnd5eapi.co")

    @cached_json
    def get_from_SRD(uri):
        return requests.get(f"{SRD_API}{uri}").json()

    return get_from_SRD


SRD = __SRD_API_CALL()

SRD_endpoints = SRD("/api/")
SRD_classes = {}
for result in SRD(SRD_endpoints["classes"])["results"]:
    SRD_classes[result["index"]] = SRD(result["url"])



from setuptools import setup, find_packages
from os import path
from dnd_character.__init__ import __version__, __author__, __credits__

try:
    with open(
        path.join(path.abspath(path.dirname(__file__)), "README.md"), encoding="utf-8"
    ) as f:
        long_description = f.read()
except Exception:
    long_description = "missing README.md"

setup(
    name="dnd_character",
    author=__author__,
    credits=__credits__,
    packages=find_packages(),
    include_package_data=True,
    version=__version__,
    url="https://github.com/tassaron2/dnd-character",
    license="EPL-2.0",
    description="make Dungeons & Dragons characters as serializable objects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="dnd trpg tabletop rpg",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Role-Playing",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
    ],
    install_requires=["requests"],
)



from uuid import uuid4, UUID
from functools import reduce
import math
import logging

from .roll import RollStats
from .SRD import SRD
from .experience import Experience, experience_at_level


LOG = logging.getLogger(__package__)


class Character:
    """
    Character Object deals with all aspects of the player character including
    name, age, gender, description, biography, level, wealth, and all
    player ability scores.  All can be omitted to create a blank, level 1
    player.
    """

    def __init__(
        self,
        uid: UUID = None,
        name: str = None,
        age: str = None,
        gender: str = None,
        alignment: str = None,
        description: str = None,
        biography: str = None,
        classs: dict = None,
        class_name: str = None,
        class_levels: list = None,
        level: int = 1,
        experience: int = 0,
        wealth: int = 0,
        strength: int = None,
        dexterity: int = None,
        constitution: int = None,
        wisdom: int = None,
        intelligence: int = None,
        charisma: int = None,
        hp: int = None,
        hd: str = None,
        proficiencies: dict = None,
        saving_throws: list = None,
        spells: dict = None,
        spells_prepared: list = None,
        spell_slots: dict = None,
        skills_strength: dict = None,
        skills_dexterity: dict = None,
        skills_wisdom: dict = None,
        skills_intelligence: dict = None,
        skills_charisma: dict = None,
        inventory: list = None,
        invsize: int = None,
        prof_bonus: int = 0,
        ability_score_bonus: int = 0,
        class_features: dict = None,
    ):
        """
        Typical Arguments:
                name         (str)
                age          (str)
                gender       (str)
                alignment    (str): two letter alignment (LE, TN, CN, LG, etc.)
                description  (str): physical description of character
                biography    (str): backstory of character
                level        (int): character's starting level
                wealth       (int): character's starting wealth
                strength     (int): character's starting strength
                dexterity    (int):  character's starting dexterity
                constitution (int):  character's starting constitution
                wisdom       (int):  character's starting wisdom
                intelligence (int):  character's starting intelligence
                charisma     (int):  character's starting charisma
                hp           (int):  character's starting hitpoint value
        """

        self.uid = UUID(uid) if uid is not None else uuid4()
        self.name = name
        self.age = age
        self.gender = gender
        self.description = description
        self.biography = biography

        self.alignment = alignment
        if self.alignment is not None:
            assert (
                len(self.alignment) == 2
            ), "Alignments must be 2 letters (i.e LE, LG, TN, NG, CN)"
            self.alignment = self.alignment.upper()

        self.wealth = wealth
        self.class_name = class_name
        self.class_levels = class_levels if class_levels is not None else []
        self.experience = Experience(character=self, experience=int(experience))
        if level != self.level:
            if self._experience._experience == 0:
                # if only level is specified, set the experience to the amount for that level
                self.experience = experience_at_level(level)
            else:
                # the Experience object normally handles the Character object's level attr
                # but if a user changes their level manually, it should override this anyway
                LOG.info(
                    f"Custom level for {str(self.name)}: {str(level)} instead of {str(self.level)}"
                )
                self.level = level

        # Ability Scores
        self.strength = self.setInitialAbilityScore(strength)
        self.dexterity = self.setInitialAbilityScore(dexterity)
        self.constitution = self.setInitialAbilityScore(constitution)
        self.wisdom = self.setInitialAbilityScore(wisdom)
        self.intelligence = self.setInitialAbilityScore(intelligence)
        self.charisma = self.setInitialAbilityScore(charisma)

        # Spells, Skills, Proficiencies
        self.hp = hp
        self.hd = hd
        self.proficiencies = proficiencies if proficiencies is not None else {}
        self.saving_throws = saving_throws if saving_throws is not None else []
        self.spells = spells
        self.spells_prepared = spells_prepared
        self.spell_slots = spell_slots
        self.skills_charisma = skills_charisma
        self.skills_wisdom = skills_wisdom
        self.skills_dexterity = skills_dexterity
        self.skills_intelligence = skills_intelligence
        self.skills_strength = skills_strength

        # Inventory (currently primitive)
        self.inventory = inventory if inventory is not None else []
        self.invsize = len(self.inventory)

        self.prof_bonus = prof_bonus
        self.ability_score_bonus = ability_score_bonus
        self.class_features = class_features if class_features is not None else {}
        self.classs = classs

    def __str__(self):
        return (
            f"Name: {self.name}\n"
            f"Age: {self.age}\n"
            f"Gender: {self.gender}\n"
            f"Description: {self.description}\n"
            f"Biography:\n{self.biography}\n\n"
            f"Class: {self.class_name}\n"
            f"Level: {str(self.level)}\n"
            f"Current Experience: {str(self.experience)}\n"
            f"EXP to next Level: {str(self.experience.to_next_level)}\n\n"
            f"Proficiencies:\n{', '.join([value['name'] for value in self.proficiencies.values()])}\n\n"
            f"Inventory:\n{', '.join([item['name'] for item in self.inventory])}\n\n"
            f"Class Features:\n{', '.join([item['name'] for item in self.class_features.values()])}\n\n"
        )

    def keys(self):
        return [key for key in self.__dict__ if not key.startswith("_")]

    def values(self):
        return [
            value if key not in ("uid", "experience") else str(value)
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        ]

    def __getitem__(self, key):
        return dict(zip(self.keys(), self.values()))[key]

    @property
    def experience(self):
        return self._experience.experience

    @experience.setter
    def experience(self, new_val):
        if new_val is None:
            pass
        elif type(new_val) is Experience:
            self._experience = new_val
        else:
            self._experience.experience = new_val

    @property
    def classs(self):
        return self.__class

    @classs.setter
    def classs(self, new_class):
        self.__class = new_class
        if new_class is not None:
            self.class_name = new_class["name"]
            self.hd = new_class["hit_die"]

            # create dict such as { "all-armor": {"name": "All armor", "type": "Armor"} }
            for proficiency in new_class["proficiencies"]:
                data = SRD(proficiency["url"])
                self.proficiencies[proficiency["index"]] = {
                    "name": data["name"],
                    "type": data["type"],
                }

            self.saving_throws = [
                saving_throw["name"] for saving_throw in new_class["saving_throws"]
            ]

            for item in SRD(new_class["starting_equipment"])["starting_equipment"]:
                for i in range(item["quantity"]):
                    self.giveItem(SRD(item["equipment"]["url"]))

            self.class_levels = SRD(new_class["class_levels"])
            self.applyClassLevel()

            return False

    def applyClassLevel(self):
        if self.level <= len(self.class_levels):
            data = self.class_levels[self.level - 1]
            self.ability_score_bonus = data.get(
                "ability_score_bonuses", self.ability_score_bonus
            )
            self.prof_bonus = data.get("prof_bonus", self.prof_bonus)
            for feat in data["features"]:
                self.class_features[feat["index"]] = SRD(feat["url"])

    # Inventory and Inventory management (Primitive)
    def getInventorySize(self):
        self.invsize = len(self.inventory)
        return self.invsize

    def updateInventory(self):
        self.invsize = len(self.inventory)

    def giveItem(self, item):
        self.inventory.append(item)
        self.updateInventory()

    def removeItem(self, item):
        self.inventory.remove(item)
        self.updateInventory()

    def giveWealth(self, amount):
        """
        Give wealth to character

        Args:
                amount

        Returns:
                None
        """
        self.wealth += amount

    def removeWealth(self, amount):
        """
        Remove wealth from character

        Args:
                amount

        Returns:
                None
        """
        if amount > self.wealth:
            self.wealth = 0
        else:
            self.wealth -= amount

    def chargeWealth(self, amount):
        """
        The chargeWealth() method functions similarly to the
        removeWealth() method. However, instead of setting
        self.wealth to 0 on the condition of not having enough,
        chargeWealth() returns False and doesn't change the
        value of self.wealth. On success, self.removeWealth()
        is triggered, the amount is removed, and chargeWealth()
        returns True.

        Args:
                amount

        Returns:
                Boolval

        Triggers:
                self.removeWealth() if self.wealth > amount
        """
        if amount > self.wealth:
            # Not enough for purchase
            return False
        else:
            self.removeWealth(self, amount)
            return True

    @classmethod
    def setInitialAbilityScore(self, stat):
        """
        Set ability score to an int. If the argument is None, then this method
        instead rolls for the initial starting ability score.

        Args:
                stat

        Returns:
                the new ability score as an int
        """
        if stat == None:
            roll = RollStats()
            return roll.stat_roller()
        else:
            return int(stat)

    @classmethod
    def getModifier(self, stat):
        """
        This method returns the modifier for the given stat

        Args:
                stat (int): The player ability score to calculate the modifier for

        Returns:
                modifier  (int): The modifier for the given stat
                None (NoneType): Returns none if the ability score queried doesn't exist
        """
        modifier = math.floor(stat / 2) - 5
        return modifier
